
#include "state.h"
#include "AI.h"

static int distanceBetween(const PieceCollection &a, const PieceCollection &b)
{
    if (a.isAdjacentTo(b)) return 1;
    if (a.numberOf(a.biggestSize()) == a.number()) return 2;
    if (b.numberOf(b.biggestSize()) == b.number()) return 2;
    if (a.numberOf(SMALL) == b.numberOf(SMALL) &&
        a.numberOf(MEDIUM) == b.numberOf(MEDIUM)) return 2;
    return 3;
}

/* Look for yellow sacrifices that cause a catastrophe in "c" at "here".
 * We break the attacker's ships up into tiers based on their distance
 * from "here". A catastrophe can be caused in any of the following ways:
 * (num_needed=3) Sacrifice y3; move three ships in from distance 1.
 * (num_needed=2) Sac y2 or y3; move two ships in from distance 1.
 * (num_needed=2) Sac y3; move one ship in from d=1 and one from d=2.
 * (num_needed=1) Sac y3; move one ship in from anywhere on the map.
 * (num_needed=1) Sac y2; move one ship in from d=1 or d=2.
 * (num_needed=1) Sac y1; move one ship in from d=1.
 * */
static int heuristic_catastrophe_penalty_yellow(const GameState &st,
                const StarSystem &here, int attacker, Color c,
                int base_cost, int worst_penalty_so_far)
{
    const int shipcost[3] = {10, 25, 50};
    const int num_needed = 4 - here.numberOf(c);
    assert(1 <= num_needed && num_needed <= 3);
    /* Figure out which of the above methods might apply. */
    int num_ships_at[4] = {0,0,0,0};
    int num_ships_ats[4][3] = {{0,0,0},{0,0,0},{0,0,0},{0,0,0}};
    int num_yellow_ships[3] = {0,0,0};
    
    for (int i=0; i < (int)st.stars.size(); ++i) {
        const StarSystem &star = st.stars[i];
        if (&star == &here) continue;
        const int d = distanceBetween(here.star, star.star);
        assert(1 <= d && d <= 3);
        num_ships_at[d] += star.ships[attacker].numberOf(c);
        num_ships_ats[d][SMALL] += star.ships[attacker].numberOf(c,SMALL);
        num_ships_ats[d][MEDIUM] += star.ships[attacker].numberOf(c,MEDIUM);
        num_ships_ats[d][LARGE] += star.ships[attacker].numberOf(c,LARGE);
        num_yellow_ships[SMALL] += star.ships[attacker].numberOf(YELLOW,SMALL);
        num_yellow_ships[MEDIUM] += star.ships[attacker].numberOf(YELLOW,MEDIUM);
        num_yellow_ships[LARGE] += star.ships[attacker].numberOf(YELLOW,LARGE);
    }
    if (num_needed == 3) {
        /* Sacrifice y3; move three ships in from distance 1 */
        if (num_yellow_ships[LARGE] == 0) return 0;
        if (num_ships_at[1] < 3) return 0;
        int smallmoved = std::min(3, num_ships_ats[1][SMALL]);
        int medmoved = std::min(3-smallmoved, num_ships_ats[1][MEDIUM]);
        int bigmoved = std::min(3-smallmoved-medmoved, num_ships_ats[1][LARGE]);
        assert(smallmoved + medmoved + bigmoved == 3);
        if (c == YELLOW) {
            assert(bigmoved <= num_yellow_ships[LARGE]);
            if (bigmoved == num_yellow_ships[LARGE]) return 0;
        }
        int penalty = base_cost + shipcost[LARGE]
                    + shipcost[LARGE]*bigmoved
                    + shipcost[MEDIUM]*medmoved
                    + shipcost[SMALL]*smallmoved;
        return std::min(penalty, worst_penalty_so_far);
    }
    if (num_needed == 2) {
        if (num_ships_at[1] >= 2 && (num_yellow_ships[LARGE] != 0 || num_yellow_ships[MEDIUM] != 0)) {
            /* Sacrifice y2 or y3; move two ships in from distance 1 */
            int smallmoved = std::min(2, num_ships_ats[1][SMALL]);
            int medmoved = std::min(2-smallmoved, num_ships_ats[1][MEDIUM]);
            int bigmoved = std::min(2-smallmoved-medmoved, num_ships_ats[1][LARGE]);
            assert(smallmoved + medmoved + bigmoved == 2);
            Size sacced = (num_yellow_ships[MEDIUM] > 0 ? MEDIUM : LARGE);
            if (c == YELLOW) {
                assert(bigmoved <= num_yellow_ships[LARGE]);
                assert(medmoved <= num_yellow_ships[MEDIUM]);
                if (medmoved == num_yellow_ships[MEDIUM]) {
                    sacced = LARGE;
                    if (bigmoved == num_yellow_ships[LARGE])
                      goto failed_alpha;
                }
            }
            int penalty = base_cost + shipcost[sacced]
                        + shipcost[LARGE]*bigmoved
                        + shipcost[MEDIUM]*medmoved
                        + shipcost[SMALL]*smallmoved;
            worst_penalty_so_far = std::min(penalty, worst_penalty_so_far);
        }
      failed_alpha:
        if (num_yellow_ships[LARGE] >= 1 &&
                 num_ships_at[2] >= 1 && num_ships_at[1] >= 1) {
            /* Sacrifice y3; move one ship in from d=1 and one from d=2. */
            int smallmoved1 = std::min(1, num_ships_ats[1][SMALL]);
            int medmoved1 = std::min(1-smallmoved1, num_ships_ats[1][MEDIUM]);
            int bigmoved1 = std::min(1-smallmoved1-medmoved1, num_ships_ats[1][LARGE]);
            int smallmoved2 = std::min(1, num_ships_ats[2][SMALL]);
            int medmoved2 = std::min(1-smallmoved2, num_ships_ats[2][MEDIUM]);
            int bigmoved2 = std::min(1-smallmoved2-medmoved2, num_ships_ats[2][LARGE]);
            assert(smallmoved1 + medmoved1 + bigmoved1 == 1);
            assert(smallmoved2 + medmoved2 + bigmoved2 == 1);
            const int smallmoved = smallmoved1 + smallmoved2;
            const int medmoved = medmoved1 + medmoved2;
            const int bigmoved = bigmoved1 + bigmoved2;
            if (c == YELLOW) {
                assert(bigmoved <= num_yellow_ships[LARGE]);
                if (bigmoved == num_yellow_ships[LARGE])
                  goto failed_beta;
            }
            int penalty = base_cost + shipcost[LARGE]
                        + shipcost[LARGE]*bigmoved
                        + shipcost[MEDIUM]*medmoved
                        + shipcost[SMALL]*smallmoved;
            worst_penalty_so_far = std::min(penalty, worst_penalty_so_far);
        }
      failed_beta:
        return worst_penalty_so_far;
    }
    assert(num_needed == 1);
    if (num_ships_at[1] >= 1) {
        /* Move one ship in from d=1. */
        int smallmoved = std::min(1, num_ships_ats[1][SMALL]);
        int medmoved = std::min(1-smallmoved, num_ships_ats[1][MEDIUM]);
        int bigmoved = std::min(1-smallmoved-medmoved, num_ships_ats[1][LARGE]);
        assert(smallmoved + medmoved + bigmoved == 1);
        Size sacced = SMALL;
        if (num_yellow_ships[SMALL] == (c==YELLOW && smallmoved ? 1 : 0)) {
            sacced = MEDIUM;
            if (num_yellow_ships[MEDIUM] == (c==YELLOW && medmoved ? 1 : 0)) {
                sacced = LARGE;
                if (num_yellow_ships[LARGE] == (c==YELLOW && bigmoved ? 1 : 0))
                  goto failed_gamma;
            }
        }
        int penalty = base_cost + shipcost[sacced]
                    + shipcost[LARGE]*bigmoved
                    + shipcost[MEDIUM]*medmoved
                    + shipcost[SMALL]*smallmoved;
        worst_penalty_so_far = std::min(penalty, worst_penalty_so_far);
    }
  failed_gamma:
    if (num_ships_at[2] >= 1) {
        /* Move one ship in from d=2. */
        int smallmoved = std::min(1, num_ships_ats[2][SMALL]);
        int medmoved = std::min(1-smallmoved, num_ships_ats[2][MEDIUM]);
        int bigmoved = std::min(1-smallmoved-medmoved, num_ships_ats[2][LARGE]);
        assert(smallmoved + medmoved + bigmoved == 1);
        Size sacced = MEDIUM;
        if (num_yellow_ships[MEDIUM] == (c==YELLOW && medmoved ? 1 : 0)) {
            sacced = LARGE;
            if (num_yellow_ships[LARGE] == (c==YELLOW && bigmoved ? 1 : 0))
              goto failed_epsilon;
        }
        int penalty = base_cost + shipcost[sacced]
                    + shipcost[LARGE]*bigmoved
                    + shipcost[MEDIUM]*medmoved
                    + shipcost[SMALL]*smallmoved;
        worst_penalty_so_far = std::min(penalty, worst_penalty_so_far);
    }
  failed_epsilon:
    if (num_ships_at[3] >= 1 && num_yellow_ships[LARGE] >= 1) {
        /* Move one ship in from d=3. */
        int smallmoved = std::min(1, num_ships_ats[3][SMALL]);
        int medmoved = std::min(1-smallmoved, num_ships_ats[3][MEDIUM]);
        int bigmoved = std::min(1-smallmoved-medmoved, num_ships_ats[3][LARGE]);
        assert(smallmoved + medmoved + bigmoved == 1);
        if (c == YELLOW && num_yellow_ships[LARGE] == num_ships_ats[3][LARGE])
          goto failed_eta;
        int penalty = base_cost + shipcost[LARGE]
                    + shipcost[LARGE]*bigmoved
                    + shipcost[MEDIUM]*medmoved
                    + shipcost[SMALL]*smallmoved;
        worst_penalty_so_far = std::min(penalty, worst_penalty_so_far);
    }
  failed_eta:
    return worst_penalty_so_far;
}


/* How convenient would it be for the defender to catastrophe color "c" at
 * this system? Look at how many ships the defender would have to lose,
 * versus how many ships the attacker would lose. Return a number that gets
 * more negative the more relative damage the defender can cause.
 * Return zero if the defender cannot catastrophe this color here, or if
 * the defender can catastrophe it only by giving up more of his own ships
 * than he causes the attacker to lose in the catastrophe.
 */
static int heuristic_catastrophe_penalty(const GameState &st,
                            const StarSystem &here, int attacker, Color c)
{
    const int shipcost[3] = {10, 25, 50};
    const int defender = 1-attacker;
    const int num_here = here.numberOf(c);
    if (num_here == 0) return 0;
    assert(1 <= num_here && num_here <= 3);
    /* Suppose all the pieces of color "c" here just vanished. What would that
     * do to the attacker's heuristic value? Things the defender would lose
     * add to the value; things the attacker would lose subtract from it. */
    int base_cost = 0;
    if (here.star.numberOf(c) == here.star.number()) {
        /* The defender is not allowed to destroy his own homeworld. */
        if (here.homeworldOf == defender) return 0;
        /* We needn't spend time on this if the defender already has a winning move. */
        if (here.homeworldOf == attacker) return 0;
        base_cost += shipcost[SMALL] * here.ships[defender].numberOf(SMALL);
        base_cost += shipcost[MEDIUM] * here.ships[defender].numberOf(MEDIUM);
        base_cost += shipcost[LARGE] * here.ships[defender].numberOf(LARGE);
        base_cost -= shipcost[SMALL] * here.ships[attacker].numberOf(SMALL);
        base_cost -= shipcost[MEDIUM] * here.ships[attacker].numberOf(MEDIUM);
        base_cost -= shipcost[LARGE] * here.ships[attacker].numberOf(LARGE);
    } else {
        base_cost += shipcost[SMALL] * here.ships[defender].numberOf(c,SMALL);
        base_cost += shipcost[MEDIUM] * here.ships[defender].numberOf(c,MEDIUM);
        base_cost += shipcost[LARGE] * here.ships[defender].numberOf(c,LARGE);
        base_cost -= shipcost[SMALL] * here.ships[attacker].numberOf(c,SMALL);
        base_cost -= shipcost[MEDIUM] * here.ships[attacker].numberOf(c,MEDIUM);
        base_cost -= shipcost[LARGE] * here.ships[attacker].numberOf(c,LARGE);
        /* Add a 100-point penalty for catastrophing part of a homeworld. */
        if (here.star.numberOf(c) != 0) {
            if (here.homeworldOf == attacker) base_cost -= 100;
            if (here.homeworldOf == defender) base_cost += 50;
        }
    }

    int worst_penalty_so_far = 0;
    /* Can we catastrophe this color here with a free move? */
    if (num_here == 3) {
        if (here.ships[defender].numberOf(c) != 0 &&
                st.stash.numberOf(c) != 0 &&
                here.playerHasAccessTo(defender, GREEN)) {
            /* The defender can just build a new ship here. */
            const int penalty = base_cost;
            if (penalty < worst_penalty_so_far)
              worst_penalty_so_far = penalty;
        } else {
            for (int i=0; i < (int)st.stars.size(); ++i) {
                const StarSystem &star = st.stars[i];
                if (!star.isAdjacentTo(here)) continue;
                if (!star.playerHasAccessTo(defender, YELLOW)) continue;
                if (star.ships[defender].numberOf(c) == 0) continue;
                if (star.homeworldOf == defender && star.ships[defender].number() == 1)
                  continue;
                /* The defender could move a ship from "star" to "here" for free,
                 * thus catastrophing "c" at "here". How much would it hurt him?
                 * He'd lose the ship he moved, plus the base cost. */
                const int penalty = base_cost + shipcost[star.ships[defender].smallestSizeOf(c)];
                if (penalty < worst_penalty_so_far)
                  worst_penalty_so_far = penalty;
            }
        }
    }
    /* Can we catastrophe this color with a sacrifice move? */
    const int num_needed = 4 - num_here;
    assert(1 <= num_needed && num_needed <= 3);
    if (here.ships[defender].numberOf(c) != 0 && 
            st.stash.numberOf(c) + (c==GREEN) >= num_needed) {
        /* If we can find a big enough green to sacrifice, we can do it.
         * Notice that if (c==GREEN), it never pays to sacrifice green at
         * the site of the catastrophe; there must be at least two of our
         * green there in order to build green back, plus at least one of
         * the opponent's green pieces, which means we could just have built
         * the fourth green for free instead of sacrificing. */
        for (Size s = (Size)(num_needed-1); s <= LARGE; ++s) {
            for (int i=0; i < (int)st.stars.size(); ++i) {
                const StarSystem &star = st.stars[i];
                if (&star == &here && c == GREEN) continue;
                if (star.ships[defender].numberOf(GREEN,s) == 0) continue;
                if (star.homeworldOf == defender && star.ships[defender].number() == 1)
                  continue;
                const int penalty = base_cost + shipcost[s];
                if (penalty < worst_penalty_so_far)
                  worst_penalty_so_far = penalty;
                goto done_green;
            }
        }
      done_green: ;
    }
    if (here.ships[defender].number() - here.ships[defender].numberOf(c) >= num_needed &&
            st.stash.numberOf(c) + (c==BLUE) >= num_needed) {
        /* If we can find a big enough blue to sacrifice, we might be able
         * to do it. We don't bother to check whether the sizes in the stash
         * actually match up with the sizes of the defender's ships;
         * TODO FIXME BUG HACK. (Honestly, it's highly unusual for someone to
         * cause a catastrophe with a blue sacrifice in the first place.) */
        for (Size s = (Size)(num_needed-1); s <= LARGE; ++s) {
            for (int i=0; i < (int)st.stars.size(); ++i) {
                const StarSystem &star = st.stars[i];
                if (star.ships[defender].numberOf(BLUE,s) == 0) continue;
                if (star.homeworldOf == defender && star.ships[defender].number() == 1)
                  continue;
                if (&star == &here && c == BLUE && here.ships[defender].number() == num_needed)
                  continue;
                const int penalty = base_cost + shipcost[s];
                if (penalty < worst_penalty_so_far)
                  worst_penalty_so_far = penalty;
                goto done_blue;
            }
        }
      done_blue: ;
    }
    /* Finally, the case that we're usually worried about: the big yellow
     * sacrifice (a.k.a. the "Bluebird Mistake"). */
    worst_penalty_so_far = heuristic_catastrophe_penalty_yellow(
            st, here, defender, c, base_cost, worst_penalty_so_far);
    /* Done! */
    assert(worst_penalty_so_far <= 0);
    return worst_penalty_so_far; 
}


/* Return an estimated value of the given position. Return a high value if
 * the attacker is winning, and a low value if he's losing. When considering
 * the stash, note that it is the defender's turn right now. */
int ai_static_evaluation(const GameState &st, int attacker)
{
    const int defender = 1-attacker;
    const StarSystem *hw[2] = { NULL, NULL };
    bool can_sacrifice_for[2][4];
    bool can_build[2][4];
    bool has_a_big[2][4];
    bool has_access_to_blue[2] = { false, false };
    int sum = 0;
    /* Fill in the can_sacrifice_for[] and has_a_big[] arrays. */
    for (int i=0; i < (int)st.stars.size(); ++i) {
        const StarSystem &star = st.stars[i];
        for (int player = 0; player <= 1; ++player) {
            if (star.playerHasAccessTo(player, BLUE))
              has_access_to_blue[player] = true;
            if (star.homeworldOf == player) {
                hw[player] = &star;
                if (star.ships[player].number() == 1)
                  continue;
            }
            for (Color c = RED; c <= BLUE; ++c) {
                if (!has_a_big[player][c] &&
                        st.stars[i].ships[attacker].numberOf(c,LARGE) != 0) {
                    has_a_big[player][c] = true;
                    can_sacrifice_for[player][c] = true;
                } else if (!can_sacrifice_for[player][c] &&
                        st.stars[i].ships[attacker].numberOf(c) != 0) {
                    can_sacrifice_for[player][c] = true;
                }
            }
        }
    }
    /* Fill in the can_build[] array. */
    for (int i=0; i < (int)st.stars.size(); ++i) {
        const StarSystem &star = st.stars[i];
        for (int player = 0; player <= 1; ++player) {
            for (Color c = RED; c <= BLUE; ++c) {
                if (can_build[player][c]) continue;
                if (star.ships[player].numberOf(c) == 0) continue;
                if (!can_sacrifice_for[player][GREEN] && !star.playerHasAccessTo(player, GREEN))
                  continue;
                /* Player can build a ship of this color here. */
                if (star.numberOf(c) == 3)
                  continue;
                /* Player can build said ship without causing a catastrophe. */
                can_build[player][c] = true;
            }
        }
    }

    assert(hw[attacker] != NULL);
    assert(hw[defender] != NULL);
    const StarSystem &attacker_hw = *hw[attacker];
    const StarSystem &defender_hw = *hw[defender];

    /* Points for having a big yellow. */
    if (has_a_big[attacker][YELLOW])
      sum += 40;
    if (has_a_big[defender][YELLOW])
      sum -= 30;
    if (!has_access_to_blue[attacker])
      sum -= 120;
    if (!has_access_to_blue[defender])
      sum += 90;

    /* The Gun Rule: Negative points for lacking red power when your
     * opponent does have red power. */
    if (can_sacrifice_for[defender][RED] && !can_sacrifice_for[attacker][RED])
      sum -= 30;
    /* Points for having access to each color. */
    if (can_sacrifice_for[attacker][RED]) sum += 20;
    if (can_sacrifice_for[attacker][YELLOW]) sum += 100;
    if (can_sacrifice_for[attacker][GREEN]) sum += 50;
    if (can_sacrifice_for[attacker][BLUE]) sum += 50;
    
    for (int i=0; i < (int)st.stars.size(); ++i) {
        const StarSystem &star = st.stars[i];
        const bool adjacent_to_defender = star.isAdjacentTo(defender_hw);
        const bool adjacent_to_attacker = star.isAdjacentTo(attacker_hw);
        if (adjacent_to_defender) {
            /* Points for having ships adjacent to the opponent's homeworld. */
            sum += 40*star.ships[attacker].numberOf(LARGE);
            sum += 27*star.ships[attacker].numberOf(MEDIUM);
            sum += 13*star.ships[attacker].numberOf(SMALL);
        } else if (adjacent_to_attacker || star.homeworldOf == attacker) {
            /* Points for having a lot of material in general. */
            sum += 50*star.ships[attacker].numberOf(LARGE);
            sum += 25*star.ships[attacker].numberOf(MEDIUM);
            sum += 10*star.ships[attacker].numberOf(SMALL);
        } else {
            /* Minor points for having ships in "far-away" systems. They
             * can be sacrificed, but that's about it. */
            sum += 30*star.ships[attacker].numberOf(LARGE);
            sum += 10*star.ships[attacker].numberOf(MEDIUM);
            sum += 2*star.ships[attacker].numberOf(SMALL);
        }
        if (adjacent_to_attacker) {
            sum -= 40*star.ships[defender].numberOf(LARGE);
            sum -= 30*star.ships[defender].numberOf(MEDIUM);
            sum -= 15*star.ships[defender].numberOf(SMALL);
        } else {
            /* Give full points for the opponent's ships even in "far-away"
             * systems; assume he has a clever plan we don't know about. */
            sum -= 50*star.ships[defender].numberOf(LARGE);
            sum -= 25*star.ships[defender].numberOf(MEDIUM);
            sum -= 10*star.ships[defender].numberOf(SMALL);
        }
        /* Points for having big greens and big yellows. */
        sum += 2*star.ships[attacker].numberOf(GREEN, LARGE);
        sum += 4*star.ships[attacker].numberOf(YELLOW, LARGE);
        sum += 1*star.ships[attacker].numberOf(GREEN, MEDIUM);
        sum += 2*star.ships[attacker].numberOf(YELLOW, MEDIUM);
        /* Points for having g1g3, g2g2, or something like that; that's
         * the "Factory", which lets you sacrifice the larger green and
         * immediately build it back. */
        if (star.ships[attacker].numberOf(GREEN) >= 2 &&
            star.ships[attacker].numberOf(GREEN) > star.ships[attacker].numberOf(GREEN, SMALL)) {
            sum += (st.stash.numberOf(GREEN,SMALL)==0 ? 15 : 10); 
        }
        if (star.ships[defender].numberOf(GREEN) >= 2 &&
            star.ships[defender].numberOf(GREEN) > star.ships[defender].numberOf(GREEN, SMALL)) {
            sum -= (st.stash.numberOf(GREEN,SMALL)==0 ? 20 : 10); 
        }
        /* Points for having the biggest ship in a contested system. */
        if (!star.ships[attacker].empty() && !star.ships[defender].empty()) {
            Size as = star.ships[attacker].biggestSize();
            Size ds = star.ships[defender].biggestSize();
            if (as > ds && (can_sacrifice_for[attacker][RED] || star.playerHasAccessTo(attacker, RED)))
                sum += 5;
            if (ds > as && (can_sacrifice_for[defender][RED] || star.playerHasAccessTo(defender, RED)))
                sum -= 10;
        }
        if (star.homeworldOf == attacker) {
            /* Points for having a large ship at home. */
            if (star.ships[attacker].numberOf(LARGE) >= 1)
              sum += 100 + 25*(star.ships[attacker].numberOf(LARGE) >= 2);
            /* Negative points for lacking colors at home. */
            if (!star.playerHasAccessTo(attacker, GREEN))
              sum -= 50;
            if (!star.playerHasAccessTo(attacker, YELLOW))
              sum -= 60;
            if (!star.playerHasAccessTo(attacker, BLUE)) {
                sum -= 60;
                /* Getting stuck with no blue at home and no blue ships to
                 * move home is a very bad position. */
                if (!can_sacrifice_for[attacker][BLUE])
                  sum -= 50;
            }
            if (star.playerHasAccessTo(attacker, RED))
              sum += 20;
            /* Having neither blue nor yellow is extremely bad! */
            if (!star.playerHasAccessTo(attacker, BLUE) &&
                !star.playerHasAccessTo(attacker, YELLOW)) sum -= 70;
            /* Points for being able to build each color at home. */
            if (star.ships[attacker].numberOf(GREEN) != 0) sum += 15;
            if (star.ships[attacker].numberOf(BLUE) != 0) sum += 10;
            if (star.ships[attacker].numberOf(YELLOW) != 0) sum += 15;
            if (star.ships[attacker].numberOf(RED) != 0) sum += 4;
            /* Letting the other player get a ship into our home is likely
             * to be disastrous, since we aren't very good at analyzing
             * that kind of situation. */
            if (!star.ships[defender].empty()) {
                sum -= 100*star.ships[defender].numberOf(LARGE);
                sum -= 70*star.ships[defender].numberOf(MEDIUM);
                sum -= 25*star.ships[defender].numberOf(SMALL);
                if (star.ships[defender].biggestSize() > star.ships[attacker].biggestSize())
                  sum -= 100;
            }
            /* Negative points for having three of a kind at home.
             * heuristic_catastrophe_penalty() takes care of the short-term
             * threat from this angle, but we want to discourage threesomes
             * in the long term, because otherwise the AI is too hesitant
             * about moving out to new systems. */
            for (Color c = RED; c <= BLUE; ++c) {
                if (star.numberOf(c) >= 3)
                  sum -= (star.star.numberOf(c) != 0 ? 60 : 20);
            }
        } else if (star.homeworldOf == defender) {
            /* Points for having a large ship at home. */
            if (star.ships[defender].numberOf(LARGE) >= 1)
              sum -= 80;
            if (!star.playerHasAccessTo(defender, GREEN)) sum += 40;
            if (!star.playerHasAccessTo(defender, YELLOW)) sum += 40;
            if (!star.playerHasAccessTo(defender, BLUE)) sum += 40;
        } else if (!star.ships[attacker].empty()) {
            /* Negative points for having ships stuck in the doldrums. */
            if (!star.playerHasAccessTo(attacker, YELLOW)) {
                if (!star.playerHasAccessTo(attacker, BLUE))
                  sum -= 5*star.ships[attacker].number();
                else
                  sum -= 2*star.ships[attacker].number();
            }
            /* Bonus for each system where you can build; bonus for
             * each system where you can use the blue power. */
            if (star.playerHasAccessTo(attacker, GREEN))
              sum += 10;
            if (star.playerHasAccessTo(attacker, BLUE))
              sum += 10;
        }
        /* Negative points for leaving enough of one color in a system that
         * the defender can produce a beneficial catastrophe on his turn. */
        if (!star.ships[attacker].empty()) {
            for (Color c = RED; c <= BLUE; ++c) {
                sum += heuristic_catastrophe_penalty(st, star, attacker, c);
            }
        }
    }
    /* Stash economics. */
    for (Color c = RED; c <= BLUE; ++c) {
        const int n = st.stash.numberOf(c);
        if (n <= 1) continue;
        const Size smallest = st.stash.smallestSizeOf(c);
        if (!can_build[defender][c]) {
            /* Points for denying a color to the defender. */
            sum += (c == RED) ? 20 : (c == BLUE) ? 40 : 75;
        } else {
            /* Leave just one small piece in the stash to prevent the
             * defender from grabbing a bigger piece for free. Note that
             * we should not value this advantage so highly that we
             * sacrifice a small ship back to the stash rather than
             * building one of two or three available mediums! */
            if (smallest != LARGE && st.stash.numberOf(c, smallest) == 1)
              sum += (can_sacrifice_for[attacker][c] ? 7 : 3);
        }
        if (!can_build[attacker][c] && smallest >= MEDIUM) {
            /* Big penalty for being locked out of a color when the defender
             * is probably building two- and three-pointers for free. */
            sum -= 30;
        }
        if (!can_build[defender][c] && smallest >= MEDIUM) {
            sum += 20;
        }
    }
    /* Penalty for having lost part of your homeworld star. */
    if (attacker_hw.star.number() == 1) sum -= 120;
    if (defender_hw.star.number() == 1) sum += 120;
    return sum;
}

