
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <algorithm>
#include <string>
#include <vector>
#include "state.h"


const StarSystem *GameState::systemNamed(const char *name) const
{
    for (int i=0; i < (int)stars.size(); ++i) {
        if (stars[i].name == name)
          return &stars[i];
    }
    return NULL;
}

StarSystem *GameState::systemNamed(const char *name)
{
    for (int i=0; i < (int)stars.size(); ++i) {
        if (stars[i].name == name)
          return &stars[i];
    }
    return NULL;
}

const StarSystem *GameState::homeworldOf(int player) const
{
    assert(0 <= player && player < NUMPLAYERS);
    for (int i=0; i < (int)stars.size(); ++i) {
        if (stars[i].homeworldOf == player)
          return &stars[i];
    }
    return NULL;
}

StarSystem *GameState::homeworldOf(int player)
{
    assert(0 <= player && player < NUMPLAYERS);
    for (int i=0; i < (int)stars.size(); ++i) {
        if (stars[i].homeworldOf == player)
          return &stars[i];
    }
    return NULL;
}

/* Returns true if there is an overpopulation of any color anywhere in
 * the galaxy. The significance of this is that there should never be any
 * overpopulations remaining at the end of a turn, if both players are
 * playing optimally; this is an invariant on which our AI relies. */
bool GameState::containsOverpopulation() const
{
    for (int i=0; i < (int)stars.size(); ++i) {
	if (stars[i].canCatastrophe(RED)) return true;
	if (stars[i].canCatastrophe(YELLOW)) return true;
	if (stars[i].canCatastrophe(GREEN)) return true;
	if (stars[i].canCatastrophe(BLUE)) return true;
    }
    return false;
}


void GameState::removeSystemNamed(const char *name)
{
    for (int i=0; i < (int)stars.size(); ++i) {
        if (stars[i].name == name) {
            stars.erase(stars.begin()+i);
            return;
        }
    }
    assert(false);
}

void GameState::removeSystem(StarSystem &star)
{
    for (int i=0; i < (int)stars.size(); ++i) {
        if (&stars[i] == &star) {
            stars.erase(stars.begin()+i);
            return;
        }
    }
    assert(false);
}


/* Set up for a two-player game. */
void GameState::newGame()
{
    stars.clear();
    stash.clear();
    for (Color c = RED; c <= BLUE; ++c) {
        for (Size s = SMALL; s <= LARGE; ++s) {
            stash.insert(c, s, 3);
        }
    }
}


bool GameState::gameIsOver() const
{
    const StarSystem *hw0 = homeworldOf(0);
    const StarSystem *hw1 = homeworldOf(1);
    if (hw0 == NULL || hw0->ships[0].empty())
      return true;
    if (hw1 == NULL || hw1->ships[1].empty())
      return true;
    return false;
}


std::string GameState::toString() const
{
    std::string result = "";
    for (int i=0; i < (int)stars.size(); ++i) {
        result += stars[i].toString();
        result += '\n';
    }
    return result;
}

/* Read a complete line from the given file and return it in "line".
 * Return true if anything was read; otherwise, return false. */
static bool readline(FILE *fp, std::string &line)
{
    line = "";
    for (int k = getc(fp); k != EOF; k = getc(fp)) {
        if (k == '\n') {
            return true;
        } else if (k == '#') {
            while (k != EOF && k != '\n')
              k = getc(fp);
            return true;
        }
        line += (char)k;
    }
    /* The file ended without a newline, or it was
     * already at end-of-file. Return TRUE if we
     * read anything from this last line. */
    return (line.length() > 0);
}

/* Trim whitespace from both ends of the given string. */
static void trim(std::string &line)
{
    int start, end;
    for (start = 0; start < (int)line.length() && isspace(line[start]); ++start)
      /* continue */ ;
    for (end = line.length()-1; end >= 0 && isspace(line[end]); --end)
      /* continue */ ;
    line.erase(end+1);
    line.erase(0, start);
}


/* Read a sequence of lines from the given file, where each line matches
 * the representation expected by StarSystem::scan()... except that blank
 * lines are ignored, and lines ending with "# comment text" are trimmed
 * before processing.
 *   Because our input is line-buffered, we read a bunch of matching lines
 * and then we read one non-matching line before returning. We don't want
 * to just throw away that line, so we return it as a std::string.
 *   If we can parse everything in the file, then we return "". This is
 * not ambiguous, because we consider a blank line (or a line consisting
 * only of a comment) to be parseable, and thus we will never return ""
 * in any other circumstance.
 */
std::string GameState::scan(FILE *fp)
{
    std::string unparsed_line = "";
    stars.clear();
    while (true) {
        std::string line;
        bool gotline = readline(fp, line);
        if (!gotline)
          break;
        trim(line);
        if (line.length() == 0)
          continue;
        StarSystem star;
        bool parsedline = star.scan(line.c_str());
        if (!parsedline) {
            unparsed_line = line;
            break;
        }
        stars.push_back(star);
    }
    stash.clear();
    for (Color c = RED; c <= BLUE; ++c) {
        for (Size s = SMALL; s <= LARGE; ++s) {
            stash.insert(c, s, NUMPLAYERS+1);
        }
    }
    for (int i=0; i < (int)stars.size(); ++i) {
        stash -= stars[i];
    }
    return unparsed_line;
}


static bool stptr_less(const std::string *a, const std::string *b)
{
    return (*a < *b);
}

std::string GameState::toComparableString() const
{
    std::string result = "";
    /* Put the homeworlds first, because they are "special" regardless of
     * their position in the galaxy or their names. */
    for (int i=0; i < NUMPLAYERS; ++i) {
        const StarSystem *hw = homeworldOf(i);
        if (hw != NULL)
          result += hw->toComparableString();
        else
          result += '!';
    }
    /* For non-homeworld stars, the only thing that matters is the piece
     * distribution of the star and ships; name and position don't matter.
     * So use StarSystem::toComparableString() to throw out the name,
     * and sort the results to get rid of position information. */
     std::vector<std::string *> v;
     for (int i=0; i < (int)stars.size(); ++i) {
         if (stars[i].homeworldOf == -1)
           v.push_back(new std::string(stars[i].toComparableString()));
     }
     std::sort(v.begin(), v.end(), stptr_less);
     for (int i=0; i < (int)v.size(); ++i) {
        result += *v[i];
        delete v[i];
     }
     return result;
 }

