
#include <algorithm>
#include <vector>
#include "state.h"
#include "StateToImage.h"
#include "ImgTrig.h"

#define SYSTEM_HEIGHT      74
#define SPACER_WIDTH       30


/* How many pixels wide is the base of a ship of size s, or the side
 * of the square representing a star of size s? */
static const int SQUARE_SIDE[3] = { 15, 29, 43 };

/* How many pixels wide is the invisible box around a ship of size s? */
static const int SHIP_WIDTH[3] = { 20, 35, 49 };
/* Around a star whose biggest piece is of size s? */
static const int STAR_WIDTH[3] = { 20, 35, 49 };
/* How much do we "kern" together two adjacent enemy ships? */
#define SHIP_KERN 5

/* How many pixels high is a ship of size s? */
static const int SHIP_HEIGHT[3] = { 22, 43, 67 };

static const unsigned char colors[4][3] = {
    { 0xFF, 0x00, 0x00 },  /* RED */
    { 0xFF, 0xFF, 0x00 },  /* YELLOW */
    { 0x00, 0xE0, 0x00 },  /* GREEN */
    { 0x00, 0x48, 0xFF },  /* BLUE */
};

struct StarSorter {
    int sorthow[3]; /* SMALL, MEDIUM, and LARGE stars sort according to this key */
    StarSorter(const GameState &st);
    bool operator()(const StarSystem &a, const StarSystem &b);
};

static void compute_layout(const std::vector<StarSystem> &stars,
                           unsigned char (*&rgb)[3], int &w, int &h, const bool doit);
static void draw_system(unsigned char (*rgb)[3], int w, int h,
                        int xpos, int ypos, const StarSystem &star);
static void draw_star(unsigned char (*rgb)[3], int w, int h,
                      int xpos, int ypos, const PieceCollection &pc);
static void draw_square(unsigned char (*rgb)[3], int w, int h,
                        int xpos, int ypos, Color color, Size size);

static void draw_ship(unsigned char (*rgb)[3], int w, int h,
                      int xpos, int ypos, int player, Color c, Size s);


/* Convert the given game state to an RGB image. */
void stateToImage(const GameState &st,
                  unsigned char (*&rgb)[3], int &w, int &h)
{
    /* We start by sorting the star systems in order from top to bottom.
     * The topmost star system is player 1's homeworld. The bottommost
     * system is player 0's homeworld. In between, we arrange the systems
     * by their distance from the worlds: a system which is adjacent to
     * player 0's homeworld but not player 1's will sort lower than a system
     * that is adjacent to neither homeworld. */
    std::vector<StarSystem> stars = st.stars;
    std::sort(stars.begin(), stars.end(), StarSorter(st));
    const int numstars = stars.size();
    assert(numstars >= 2);
    assert(stars[0].homeworldOf == 1);
    assert(stars[numstars-1].homeworldOf == 0);
    /* Compute the width and height of the final image. */
    compute_layout(stars, rgb, w, h, /*doit=*/false);
    assert(w > 0 && h > 0);
    assert(w < INT_MAX/h);
    /* Allocate the buffer for the final image. */
    rgb = (unsigned char (*)[3])calloc(w*h, sizeof *rgb);
    assert(rgb != NULL);
    /* Finally, draw the image into that buffer. */
    compute_layout(stars, rgb, w, h, /*doit=*/true);
}

/* Now arrange the star systems in rows. Two star systems may go in the
 * same row if they are in the same "bucket" in terms of distance from
 * the homeworlds, and if their combined widths are less than five times
 * SYSTEM_HEIGHT (an arbitrary number, in order to keep the playing
 * field from getting ridiculously wide). Star systems in the same row
 * are separated by SPACER_WIDTH pixels. */
static void compute_layout(const std::vector<StarSystem> &stars,
                           unsigned char (*&rgb)[3], int &w, int &h, const bool doit)
{
    int current_row_width = 0;
    int current_image_height = 0;
    int max_row_width = 0;
    const int numstars = (int)stars.size();
    assert(numstars >= 2);
    if (!doit) {
        /* The first time through, we're just figuring out the values of "w"
         * and "h" to pass back to the caller. */
        w = 0;
        h = 0;
    } else {
        /* The second time, we don't write to "w" or "h"; we actually fill in
         * the pixels of the image. */
        assert(rgb != NULL);
        assert(w > 0);
        assert(h > 0);
    }

    for (int i=0; i < numstars; ++i) {
        int this_system_width = STAR_WIDTH[stars[i].star.biggestSize()];
        for (int player = 0; player < NUMPLAYERS; ++player) {
            for (Color c = RED; c <= BLUE; ++c) {
                for (Size s = SMALL; s <= LARGE; ++s) {
                    int n = stars[i].ships[player].numberOf(c,s);
                    this_system_width += (n * SHIP_WIDTH[s]);
                }
            }
        }
        /* We have now computed exactly how many pixels wide (and high)
         * the box containing this star system must be. Now, do we fit
         * this box onto the same row as the last box, or do we make a
         * new box?
         */
        assert((i==0) == (current_row_width==0));
        if (i==0 ||
                stars[i].star.biggestSize() != stars[i-1].star.biggestSize() ||
                stars[i].homeworldOf != stars[i-1].homeworldOf ||
                current_row_width+SPACER_WIDTH+this_system_width > 5*SYSTEM_HEIGHT) {
            /* This star must start a new row. */
            if (current_row_width > max_row_width)
              max_row_width = current_row_width;
            current_image_height += SYSTEM_HEIGHT;
            current_row_width = 0;
        } else {
            /* This star can go on the end of the current row. */
            current_row_width += SPACER_WIDTH;
        }
        if (doit) {
            /* Actually draw the stuff in this system. */
            printf("Drawing system %s\n", stars[i].toString().c_str());
            draw_system(rgb, w, h, current_row_width, current_image_height, stars[i]);
        }
        current_row_width += this_system_width;
    }
    
    /* Maybe the last row is the widest row. */
    if (current_row_width > max_row_width)
      max_row_width = current_row_width;
    
    if (!doit) {
        /* The first time through, we're just figuring out the values of "w"
         * and "h" to pass back to the caller. */
        w = max_row_width;
        h = current_image_height;
    }
}


static void draw_system(unsigned char (*rgb)[3], int w, int h,
                        int xpos, int ypos, const StarSystem &star)
{
    const int baseline = ypos - (SYSTEM_HEIGHT/2);
    const int star_width = STAR_WIDTH[star.star.biggestSize()];
    draw_star(rgb,w,h, xpos + star_width/2, baseline, star.star);
    /* Now draw the individual ships. */
    int xposition = xpos + star_width;
    bool something_drawn = false;
    for (int player = 0; player < NUMPLAYERS; ++player) {
        for (Size s2 = SMALL; s2 <= LARGE; ++s2) {
            /* Show ships from large to small (left to right). */
            Size s = (Size)((int)LARGE + (int)SMALL - (int)s2);
            for (Color c = RED; c <= BLUE; ++c) {
                int n = star.ships[player].numberOf(c,s);
                for (int j=0; j < n; ++j) {
                    assert(xposition < w);
                    draw_ship(rgb,w,h, xposition, baseline, player,c,s);
                    xposition += SHIP_WIDTH[s];
                    assert(xposition <= w);
                    something_drawn = true;
                }
            }
        }
        if (something_drawn)
          xposition -= SHIP_KERN;
    }
}

/* xpos is the x-coordinate of the center of the square.
 * ypos is the y-coordinate of the center of the square.
 */
static void draw_star(unsigned char (*rgb)[3], int w, int h,
                      int xpos, int ypos, const PieceCollection &pc)
{
    assert(pc.number() == 1 || pc.number() == 2);
    const Size bigsize = pc.biggestSize();
    if (pc.number() == 1) {
        /* The star consists of a single piece. Easy. */
        for (Color c = RED; c <= BLUE; ++c) {
            if (pc.numberOf(c, bigsize) == 0) continue;
            draw_square(rgb,w,h, xpos, ypos, c, bigsize);
        }
    } else if (pc.numberOf(bigsize) == 2) {
        /* The star consists of two pieces of the same size. Paint one
         * on top of the other, offset up and left by a few pixels. */
        bool first = true;
        for (Color c = RED; c <= BLUE; ++c) {
            for (int i=0; i < pc.numberOf(c, bigsize); ++i) {
                if (first) {
                    draw_square(rgb,w,h, xpos-3, ypos-3, c, bigsize);
                } else {
                    draw_square(rgb,w,h, xpos+3, ypos+3, c, bigsize);
                }
            }
        }
    } else {
        /* The star consists of two pieces of different sizes. Paint
         * the bigger one first, then the smaller one. */
        assert(bigsize != SMALL);
        for (Color c = RED; c <= BLUE; ++c) {
            if (pc.numberOf(c, bigsize) > 0) {
                draw_square(rgb,w,h, xpos,ypos, c, bigsize);
                break;
            }
        }
        /* Now draw the smaller one. */
        for (Color c = RED; c <= BLUE; ++c) {
            if (pc.numberOf(c) > pc.numberOf(c,bigsize)) {
                draw_square(rgb,w,h, xpos,ypos, c, pc.smallestSizeOf(c));
                break;
            }
        }
    }
}

static void draw_square(unsigned char (*rgb)[3], int w, int h,
                        int xpos, int ypos, Color c, Size s)
{
    assert(SMALL <= s && s <= LARGE);
    assert(RED <= c && c <= BLUE);
    assert(0 < xpos && xpos < w);
    assert(0 < ypos && ypos < h);

    const int side = SQUARE_SIDE[s];
    
    /* Draw the square. */
    for (int i=0; i < side; ++i) {
        const int y = ypos - (side/2) + i;
        for (int j=0; j < side; ++j) {
            const int x = xpos - (side/2) + j;
            memcpy(rgb[y*w+x], colors[c], 3);
        }
    }
    /* Draw the tick marks along its lower edge. */
    const int ticks = 1 + (int)s;
    for (int i=0; i < ticks; ++i) {
        const int x = xpos + (side/2) - 4*i - 4;
        const int y = ypos + (side/2) - 2;
        memset(rgb[(y-0)*w+(x-0)], '\0', 3);
        memset(rgb[(y-0)*w+(x-1)], '\0', 3);
        memset(rgb[(y-1)*w+(x-0)], '\0', 3);
        memset(rgb[(y-1)*w+(x-1)], '\0', 3);
        memset(rgb[(y-2)*w+(x-0)], '\0', 3);
        memset(rgb[(y-2)*w+(x-1)], '\0', 3);
    }
}


namespace DrawShipPlotter {
    unsigned char (*rgb)[3];
    int w, h;
    unsigned char color[3];

    int scanline(int y, int xleft, int xright)
    {
        /* Triangles drawn by draw_ship() will always be in bounds. */
        assert(0 <= y && y < h);
        assert(0 <= xleft && xleft <= xright && xright < w);
        for (int i=xleft; i <= xright; ++i)
          memcpy(rgb[y*w+i], color, 3);
        return 0;
    }
}

/* "xpos" is the left edge of the box containing the ship.
 * "ypos" is the baseline where the middle of the ship should go.
 */ 
static void draw_ship(unsigned char (*rgb)[3], int w, int h,
                      int xpos, int ypos, int player, Color c, Size s)
{
    DrawShipPlotter::rgb = rgb;
    DrawShipPlotter::w = w;
    DrawShipPlotter::h = h;
    memcpy(DrawShipPlotter::color, colors[c], 3);
    const int playerdir = (player==0) ? -1 : +1;
    const int shipcenter_x = xpos + SHIP_WIDTH[s]/2;
    const int shipleft_x = shipcenter_x - SQUARE_SIDE[s]/2;
    const int shipright_x = shipcenter_x + SQUARE_SIDE[s]/2;
    const int shiptip_y = ypos + playerdir*SHIP_HEIGHT[s]/2;
    const int shipbase_y = shiptip_y - playerdir*SHIP_HEIGHT[s];
    /* Draw the triangle. */
    plot_triangle(shipcenter_x, shiptip_y,
                  shipleft_x, shipbase_y,
                  shipright_x, shipbase_y,
                  DrawShipPlotter::scanline);
    /* Draw the tick marks along its lower edge. */
    const int ticks = 1 + (int)s;
    const int tick_y = shipbase_y + playerdir*3 + (player==1 ? 4 : 0);
    for (int i=0; i < ticks; ++i) {
        const int x = shipright_x - 4*i - 6;
        const int y = tick_y;
        memset(rgb[(y-0)*w+(x-0)], '\0', 3);
        memset(rgb[(y-0)*w+(x-1)], '\0', 3);
        memset(rgb[(y-1)*w+(x-0)], '\0', 3);
        memset(rgb[(y-1)*w+(x-1)], '\0', 3);
        memset(rgb[(y-2)*w+(x-0)], '\0', 3);
        memset(rgb[(y-2)*w+(x-1)], '\0', 3);
        memset(rgb[(y-3)*w+(x-0)], '\0', 3);
        memset(rgb[(y-3)*w+(x-1)], '\0', 3);
    }
}



StarSorter::StarSorter(const GameState &st)
{
    assert(NUMPLAYERS == 2);
    const StarSystem *hw0 = st.homeworldOf(0);
    const StarSystem *hw1 = st.homeworldOf(1);
    assert(hw0 != NULL);
    assert(hw1 != NULL);
    for (Size s = SMALL; s <= LARGE; ++s) {
        const bool adjacent_to_bottom = (hw0->star.numberOf(s) == 0);
        const bool adjacent_to_top = (hw1->star.numberOf(s) == 0);
        if (adjacent_to_bottom == adjacent_to_top) {
            sorthow[s] = 2;
        } else if (adjacent_to_bottom) {
            assert(adjacent_to_bottom && !adjacent_to_top);
            sorthow[s] = 3;
        } else {
            assert(!adjacent_to_bottom && adjacent_to_top);
            sorthow[s] = 1;
        }
    }
}

/* This function compares its arguments and returns "true" if system "a"
 * should be sorted farther toward the bottom of the image than system "b".
 */
bool StarSorter::operator()(const StarSystem &a, const StarSystem &b)
{
    if (a.homeworldOf == 1 || b.homeworldOf == 0) return true;
    if (a.homeworldOf == 0 || b.homeworldOf == 1) return false;
    assert(a.homeworldOf == -1 && b.homeworldOf == -1);
    assert(a.star.number() == 1);
    assert(b.star.number() == 1);
    /* First, sort stars by their adjacencies to the homeworlds. */
    const int asort = sorthow[a.star.biggestSize()];
    const int bsort = sorthow[b.star.biggestSize()];
    if (asort != bsort)
      return (asort < bsort);
    /* These two stars are in the same equivalence class as far as distance
     * from either homeworld is concerned. Sort a system whose ships all
     * belong to player 0 below a system whose ships all belong to player 1.
     * TODO FIXME BUG HACK: See if that is worth implementing. For now,
     * just consider these two systems "equal" in sorting order.
     */
    return false;
}

