
/* Try to play a game of Homeworlds against the human opponent. */

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "getline.h"
#include <string>
#include <vector>
#include <list>
#include "state.h"
#include "move.h"
#include "AllMoves.h"
#include "ApplyMove.h"
#include "PlanetNames.h"

/* Player numbers. The human player is player 0; the AI is player 1.
 * But this doesn't imply that the human always goes first. */
#define HUMAN 0
#define COMPUTER 1

#define randint0(n) (rand() % (n))
#define do_crash do_error


static std::list<GameState> history;


static void do_error(const char *msg, ...)
{
    va_list ap;
    va_start(ap, msg);
    printf("Error: ");
    vprintf(msg, ap);
    putchar('\n');
    va_end(ap);
    exit(EXIT_FAILURE);
}


static void display_new_state(const GameState &st, bool gameover)
{
    printf("--------------------------------%s\n",
                 (gameover ? "GAME OVER" : "---------"));
    printf("%s", st.toString().c_str());
    printf("-----------------------------------------\n");
}

static void update_transcripts(const GameState &st, const WholeMove &move,
                               const char *playername, int attacker)
{
    static int moveno = 0;
    moveno += 1;
    /* Log (append) the AI's move into "transcript.txt". */
    FILE *transcript_txt = fopen("transcript.txt", "a");
    /* Ignore filesystem errors for the sake of simplicity. TODO FIXME BUG HACK */
    assert(transcript_txt != NULL);
    fprintf(transcript_txt, "%d. %s (%d): %s\n", moveno, playername, attacker, move.toString().c_str());
    fclose(transcript_txt);
}

static void do_help()
{
    puts("A \"move\" is a sequence of \"actions\", strung together with semicolons.");
    puts("For example: \"catastrophe red at Alpha; move b3 from Beta to Alpha\".");
    puts("Recognized actions are:");
    puts("  catastrophe <color> at <name>");
    puts("  sacrifice <piece> at <name>");
    puts("  pass");
    puts("  capture <piece> at <name>");
    puts("  move <piece> from <name> to <name>");
    puts("  move <piece> from <name> to <name> (<piece>)");
    puts("  build <piece> at <name>");
    puts("  convert <piece> to <piece> at <name>");
    puts("For example, to create a new star system named \"Gamma\", you might enter");
    puts("\"move r1 from Alpha to Gamma (b2)\". This would create a new system named");
    puts("\"Gamma\", with a medium blue star.");
    puts("The input parser is case-sensitive and whitespace-sensitive.");
    return;
}

static void setup_human(GameState &st)
{
    printf("Enter a name for your homeworld. > "); fflush(stdout);
    char *moveline = NULL;
  get_name:
    char *result = getline_113(&moveline);
    /* Ignore input errors for the sake of simplicity. TODO FIXME BUG HACK */
    assert(result == moveline && result != NULL);
    if (!StarSystem::is_valid_name(moveline)) {
        printf("The name must be valid (essentially, alphanumeric). \"%s\" is not a valid name.\n",
            moveline);
        printf("Please enter a name for your homeworld. > "); fflush(stdout);
        free(moveline);
        goto get_name;
    }
    assert(moveline != NULL);
    assert(StarSystem::is_valid_name(moveline));
    st.stars.push_back(StarSystem(moveline));
    StarSystem &hw = st.stars.back();
    hw.homeworldOf = HUMAN;
    free(moveline);

    printf("Enter the initial pieces for your star. For example, \"y3b2\". > "); fflush(stdout);
  get_star_pieces:
    result = getline_113(&moveline);
    /* Ignore input errors for the sake of simplicity. TODO FIXME BUG HACK */
    assert(result == moveline && result != NULL);
    PieceCollection pc;
    bool success = pc.scan(moveline);
    if (!success) {
        printf("Your input did not appear to be a valid piece collection.\n");
        printf("A piece collection consists of alternating colors (rygb) and sizes (123),\n");
        printf("case-sensitive, with no intervening spaces or punctuation. For example,\n");
        printf("\"y3b2\" or \"r1r2\" or \"g1b1\". You entered: \"%s\"\n", moveline);
        printf("Please enter a valid piece collection for your star. > "); fflush(stdout);
        free(moveline);
        goto get_star_pieces;
    }
    /* The PieceCollection was syntactically valid, but in order to be semantically valid
     * it must consist of exactly two pieces, and those pieces must be available in the stash. */
    if (pc.number() != 2) {
        printf("Your star must contain two pieces, not %d.\n", pc.number());
        printf("Please re-enter the initial pieces for your star. > "); fflush(stdout);
        free(moveline);
        goto get_star_pieces;
    }
    if (!st.stash.contains(pc)) {
        /* Note that the only way for the stash not to contain a collection of
         * only two pieces is for the AI player to have built first --- and to
         * have picked two pieces of the same color and size, which is highly
         * unlikely to be a good starting move!  A pristine stash contains
         * three pieces of each denomination. */
        printf("Your star must consist of pieces available in the stash.\n");
        printf("(Remember, your opponent built first.) The stash contains:\n");
        printf("%s\n", st.stash.toString().c_str());
        printf("Please re-enter the initial pieces for your star. > "); fflush(stdout);
        free(moveline);
        goto get_star_pieces;
    }
    free(moveline);
    hw.star = pc;
    st.stash -= pc;

    printf("Enter your starting ship. For example, \"g3\". > "); fflush(stdout);
  get_ship:
    result = getline_113(&moveline);
    /* Ignore input errors for the sake of simplicity. TODO FIXME BUG HACK */
    assert(result == moveline && result != NULL);
    success = pc.scan(moveline);
    if (!success) {
        printf("Your input did not appear to be a valid piece collection.\n");
        printf("A piece collection consists of alternating colors (rygb) and sizes (123),\n");
        printf("case-sensitive, with no intervening spaces or punctuation. For example,\n");
        printf("\"y3\" or \"r2\" or \"g1\". You entered: \"%s\"\n", moveline);
        printf("Please enter a valid piece collection for your ship. > "); fflush(stdout);
        free(moveline);
        goto get_ship;
    }
    /* The PieceCollection was syntactically valid, but in order to be semantically valid
     * it must consist of exactly one piece, and that piece must be available in the stash. */
    if (pc.number() != 1) {
        printf("You must pick one starting ship, not %d.\n", pc.number());
        printf("Please re-enter your starting ship. > "); fflush(stdout);
        free(moveline);
        goto get_ship;
    }
    if (!st.stash.contains(pc)) {
        /* Note that the only way for the stash not to contain a collection of
         * only three pieces is for the AI player to have built first --- and to
         * have picked two pieces of the same color and size, which is highly
         * unlikely to be a good starting move!  A pristine stash contains
         * three pieces of each denomination. */
        printf("You must pick a ship from the pieces available in the stash.\n");
        printf("(Remember, your opponent built first.) The stash contains:\n");
        printf("%s\n", st.stash.toString().c_str());
        printf("Please re-enter your starting ship. > "); fflush(stdout);
        free(moveline);
        goto get_ship;
    }
    free(moveline);
    hw.ships[HUMAN] = pc;
    st.stash -= pc;
    return;
}


static GameState *human_move_and_record(const GameState &st)
{
    /* The human will choose a move leading to a new state. Push that new state
     * onto the end of the "history" list. */
    history.push_back(st);
    GameState &newst = history.back();
    
    char *moveline = NULL;
  get_move:
    /* Prompt for the human's move, and read in a line. */
    printf("Your move? > "); fflush(stdout);
    char *result = getline_113(&moveline);
    /* Ignore input errors for the sake of simplicity. TODO FIXME BUG HACK */
    assert(result == moveline && result != NULL);
    
    if (strcmp(moveline, "help") == 0) {
        do_help();
        free(moveline);
        goto get_move;
    }
    
    WholeMove move;
    bool success = move.scan(moveline);
    if (!success) {
        printf("The given string did not parse as a move. It was:\n");
        printf("\"%s\"\n", moveline);
        printf("Enter the string \"help\" for help with this game's interface.\n");
        free(moveline);
        goto get_move;
    }
    /* If we've gotten this far, the user gave us a syntactically correct move.
     * Try to apply it; if it's semantically invalid or illegal, reject it. */
    assert(success);
    free(moveline);
    success = ApplyMove::Whole(newst, HUMAN, move);
    if (!success) {
        printf("The move as parsed was invalid or disallowed by the rules. It was:\n");
        printf("\"%s\"\n", move.toString().c_str());
        printf("Enter the string \"help\" for help with this game's interface.\n");
        newst = st;
        goto get_move;
    } else if (newst.homeworldOf(HUMAN) == NULL) {
        printf("You have forfeited.\n");
    } else {
        /* Acknowledge the move's validity before printing the new game state. */
        printf("Okay.\n");
    }
    const bool game_over = (newst.homeworldOf(COMPUTER) == NULL) || (newst.homeworldOf(HUMAN) == NULL);
    display_new_state(newst, game_over);
    update_transcripts(newst, move, "human", HUMAN);
    return &newst;
}


static void setup_ai(GameState &st)
{
    StarSystem hw("Delos");
    /* Sneaky humans will try to name their world "Delos" to trick us. */
    if (st.systemNamed("Delos") != NULL)
      hw.name = "Magrathea";
    hw.homeworldOf = COMPUTER;
    do {
        hw.star.clear();
        hw.ships[COMPUTER].clear();
        /* The star should contain two colors, including at least one yellow or blue. */
        const Color c1 = (randint0(2)==0 ? YELLOW : BLUE);
        const Color c2 = Color((1+randint0(3)+(int)c1) % 4);
        /* The star should contain two different sizes. */
        const Size s1 = Size(randint0(3));
        const Size s2 = Size((s1+1+randint0(2)) % 3);
        /* The initial ship should be big and non-red. If the star doesn't have green
         * already, then the initial ship must be green. Otherwise, the ship must be
         * of a color that's not already in the star itself. */
        const Color shipc = (c2 != GREEN) ? GREEN : (c1 == YELLOW) ? BLUE : YELLOW;
        /* If this configuration isn't actually possible, rinse and repeat. */
        hw.star.insert(c1, s1);
        hw.star.insert(c2, s2);
        hw.ships[COMPUTER].insert(shipc, LARGE);
    } while (!st.stash.contains(hw.pieceCollection()));
    st.stash -= hw.star;
    st.stash -= hw.ships[COMPUTER];
    st.stars.push_back(hw);
    printf("AI picked \"%s\"\n", hw.name.c_str());
    return;
}

enum GoalKind {
    GK_DEFAULT
};

static GoalKind ai_pick_goal(const GameState &st)
{
    return GK_DEFAULT;
}

static int heuristic_value_of(const GameState &st, int who, GoalKind goal)
{
    /* The more ships I have, the better.
     * The more colors I have at each star, the better.
     * The more big ships I have, the better.
     * The more ships I have within one hop of your homeworld, the better.
     */
    struct HV {
        int ships;
        int stars;       /* stars where I have any ships */
        int per_color_per_star;
        int medships;
        int bigships;
        int closeships;  /* ships close to your homeworld */
        int atships;     /* ships at your homeworld */
    } sums[2] = {0};
    
    const StarSystem *homeworlds[2] = { st.homeworldOf(0), st.homeworldOf(1) };
    
    for (int i=0; i < st.stars.size(); ++i) {
        const StarSystem &star = st.stars[i];
        for (int p=0; p < NUMPLAYERS; ++p) {
            const int nships = star.ships[p].number();
            sums[p].ships += nships;
            sums[p].stars += (nships > 0);
            sums[p].medships += star.ships[p].numberOf(MEDIUM);
            sums[p].bigships += star.ships[p].numberOf(LARGE);
            if (&star == homeworlds[1-p])
              sums[p].atships += nships;
            else if (star.isAdjacentTo(*homeworlds[1-p]))
              sums[p].closeships += nships;
            if (nships == 1) {
                sums[p].per_color_per_star += 1;
            } else if (nships != 0) {
                for (Color c = RED; c <= BLUE; ++c)
                  sums[p].per_color_per_star += (star.ships[p].numberOf(c) != 0);
            }
        }
    }
    int value_to_me =  1*sums[who].ships
                     + 1*(sums[who].per_color_per_star / sums[who].stars)
                     + 2*sums[who].medships
                     + 5*sums[who].bigships
                     + 2*sums[who].closeships
                     + 2*sums[who].atships;
    int value_to_you = 1*sums[1-who].ships
                     + 1*(sums[1-who].per_color_per_star / sums[1-who].stars)
                     + 2*sums[1-who].medships
                     + 5*sums[1-who].bigships
                     + 2*sums[1-who].closeships
                     + 2*sums[1-who].atships;
    if (5000 + value_to_me - value_to_you < 1) {
        return 1;
    } else {
        return (5000 + value_to_me - value_to_you);
    }
}

/* Given all the possible moves from this state, pick the one that leads to the
 * new state with the highest value. */
static WholeMove *ai_pick_move(const GameState &st, std::vector<WholeMove> &allmoves)
{
    GoalKind current_goal = ai_pick_goal(st);
    GameState newst;
    WholeMove *best = NULL;
    int maxvalue = -1;
    printf("AI found %d possible moves.\n", (int)allmoves.size());
    for (int i=0; i < allmoves.size(); ++i) {
        WholeMove *cand = &allmoves[i];
        newst = st;
        printf("Evaluating \"%s\"...\n", cand->toString().c_str());
        ApplyMove::or_die(newst, COMPUTER, *cand);
        int value = heuristic_value_of(newst, COMPUTER, current_goal);
        printf("  Value was %d\n", value);
        if (value > maxvalue) {
            best = cand;
            maxvalue = value;
        }
    }
    return best;
}

static const char *ai_system_names[21] = {
    "Aleph", "Beth", "Capek", "Daleth", "Femto", "Gimel", "Hoth", "Imogen",
    "Jetta", "Kale", "Lamech", "Mandrake", "Nix", "Otto", "Pesach", "Quorum",
    "Resh", "Samael", "Tau", "Upsilon", "Easteregg"
};

static GameState *ai_move_and_record(const GameState &st)
{
    /* The AI will choose a move leading to a new state. Push that new state
     * onto the end of the "history" list. */
    history.push_back(st);
    GameState &newst = history.back();

    std::vector<WholeMove> allmoves;
    std::vector<WholeMove> winmoves;
    findAllMoves(st, COMPUTER, &winmoves, &allmoves);
    const int an = allmoves.size();
    const int wn = winmoves.size();
    const bool was_winning_move = (wn > 0);
    assert(an > 0);  /* because "pass" is always a legal move */
    assert(wn <= an);
    WholeMove *chosenmove = NULL;
    if (was_winning_move) {
        chosenmove = &winmoves[0];
    } else {
        chosenmove = ai_pick_move(st, allmoves);
    }
    assert(chosenmove != NULL);
    PlanetNames::reassign(*chosenmove, st, ai_system_names);
    printf("The AI chose the following %smove:\n",
        was_winning_move ? "winning " : "");
    puts(chosenmove->toString().c_str());
    newst = st;
    ApplyMove::or_die(newst, COMPUTER, *chosenmove);
    display_new_state(newst, was_winning_move);
    update_transcripts(newst, *chosenmove, "AI", COMPUTER);
    return &newst;
}

int main(void)
{
    srand(0);
    history.clear();
    history.push_back(GameState());
    GameState *st = &history.back();
    /* Set up the stash for a two-player game. */
    for (Color c = RED; c <= BLUE; ++c) {
        for (Size s = SMALL; s <= LARGE; ++s) {
            st->stash.insert(c,s,3);
        }
    }
    /* Flip a coin to see who sets up first. The player who sets up first
     * will move second; the player who sets up second will move first. */
    if (randint0(2) == 0) {
        printf("Human will set up first and move second.\n");
        setup_human(*st);
        setup_ai(*st);
        display_new_state(*st, false);
        st = ai_move_and_record(*st);
    } else {
        setup_ai(*st);
        display_new_state(*st, false);
        printf("Human will move first.\n");
        setup_human(*st);
        display_new_state(*st, false);
    }
    while (1) {
        st = human_move_and_record(*st);
        if (st->homeworldOf(HUMAN) == NULL) {
            /* A human player can deliberately lose the game.
             * Note that if both homeworlds are destroyed simultaneously,
             * then the player doing the destroying loses by default. */
            printf("Game over -- Computer wins by default!\n");
            break;
        } else if (st->homeworldOf(COMPUTER) == NULL) {
            printf("Game over -- Human wins!\n");
            break;
        }
        st = ai_move_and_record(*st);
        assert(st->homeworldOf(COMPUTER) != NULL);
        if (st->homeworldOf(HUMAN) == NULL) {
            printf("Game over -- Computer wins!\n");
            break;
        }
    }
    return 0;
}

