
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <vector>
#include "state.h"
#include "move.h"
#include "ApplyMove.h"
#include "AllMoves.h"

#include "ImageFmtc.h"



/* Read a complete line from the given file and return it in "line".
 * Return true if anything was read; otherwise, return false. */
static bool readline(FILE *fp, std::string &line)
{
    line = "";
    for (int k = getc(fp); k != EOF; k = getc(fp)) {
        if (k == '\n') {
            return true;
        } else if (k == '#') {
            while (k != EOF && k != '\n')
              k = getc(fp);
            return true;
        }
        line += (char)k;
    }
    /* The file ended without a newline, or it was
     * already at end-of-file. Return TRUE if we
     * read anything from this last line. */
    return (line.length() > 0);
}

struct Point {
    int gameno;
    int moveno;
    int winner_moves;
    int loser_moves;
    Point(int g, int m, int wn, int ln): gameno(g), moveno(m),
        winner_moves(wn), loser_moves(ln) { }
};

static std::vector<Point> allpoints;
static int max_moveno = 0;
static int max_winnermoves = 10;
static FILE *pointsfile = NULL;

void process(FILE *in, int gameno)
{
    GameState st;
    std::string line = st.scan(in);
    int attacker = 0;
    int moveno = 0;
    int moves_last_turn = 0;
    while (1) {
        std::vector<WholeMove> winmoves;
        std::vector<WholeMove> allmoves;
        findAllMoves(st, attacker, &winmoves, &allmoves);
        const int moves_this_turn = allmoves.size();
        printf("%d %4d %6d%s\n", moveno, (int)winmoves.size(), moves_this_turn,
            (winmoves.size() ? "    !!!!!!" : ""));
        fprintf(pointsfile, "%d %d %d %d %d\n", gameno, moveno, moves_this_turn, (int)winmoves.size(), moves_last_turn);
        const int winner_moves = std::max(moves_this_turn, moves_last_turn);
        const int loser_moves = std::min(moves_this_turn, moves_last_turn);
        allpoints.push_back(Point(gameno, moveno, winner_moves, loser_moves));
        if (moveno > max_moveno)
          max_moveno = moveno;
        if (winner_moves > max_winnermoves)
          max_winnermoves = winner_moves;
        moves_last_turn = moves_this_turn;
        /* Now advance the game state. */
        printf("Scanning move \"%s\"\n", line.c_str());
        WholeMove move(line);
        ApplyMove::or_die(st, attacker, move);
        attacker = 1-attacker;
        ++moveno;
      get_another_line:
        if (!readline(in, line)) {
            /* End-of-file. */
            break;
        }
        if (line == "") {
            /* It was a comment or blank line. */
            goto get_another_line;
        }
    }
}

/* Given h in [0.,360.) and s,v in [0.,1.], produce r,g,b in [0,255]. */
static void hsv2rgb(double h, double s, double v, int &r, int &g, int &b)
{
    int hi = (int)(h / 60.) % 6;
    double f = (h / 60.) - (int)(h / 60.);
    int px = (int)(255. * v * (1. - s));
    int qx = (int)(255. * v * (1. - f*s));
    int tx = (int)(255. * v * (1. - (1.-f)*s));
    int vx = (int)(255. * v);
    switch (hi) {
        case 0: r = vx; g = tx; b = px; break;
        case 1: r = qx; g = vx; b = px; break;
        case 2: r = px; g = vx; b = tx; break;
        case 3: r = px; g = qx; b = vx; break;
        case 4: r = tx; g = px; b = vx; break;
        case 5: r = vx; g = px; b = qx; break;
    }
}

static void game2color(int gameno, int Ngames, unsigned char *rgb)
{
    double h = 360.*gameno / Ngames;
    int r=0, g=0, b=0;
    hsv2rgb(h, 0.9, 1.0, r, g, b);
    rgb[0] = std::max(0, std::min(r, 255));
    rgb[1] = std::max(0, std::min(g, 255));
    rgb[2] = std::max(0, std::min(b, 255));
}

static int yscale(int y)
{
    /* y >= 0,
     * so y+11 >= 11,
     * so log(y+11) >= log(11) >= 1.0,
     * so 20.0*log(y+11) >= 20.0,
     * so subtracting 20.0 leaves us >= 0.0
     */
    return 20.0*log(y+11) - 20.0;
}

/* Construct a nice image from "allpoints". */
void create_image(const char *filename, int N)
{
    /* We have processed N games. We'll graph each point as a colored bar
     * at x-coordinate (gameno*N+moveno), stretching vertically from
     * scale(winner_moves) to scale(loser_moves), colored according to
     * gameno. */
    int w = (max_moveno * N) + 10;
    int h = yscale(max_winnermoves) + 10;
    unsigned char (*im)[3] = (unsigned char (*)[3])calloc(w, h * sizeof *im);
    if (im == NULL) {
        printf("Couldn't allocate memory for the final %dx%d image\n", w, h);
        return;
    }
    printf("Allocated %dx%d\n", w, h);

    /* Draw red lines at logarithmic scale. */
    for (int yy = 1; ; yy *= 10) {
        int line = h+5 - yscale(yy);
        if (line < 0 || h <= line) break;
        for (int i=0; i < w; ++i)
          im[line*w+i][0] = 255;
    }
    
    /* Plot the actual bars. */
    for (int i=0; i < (int)allpoints.size(); ++i) {
        const int g = allpoints[i].gameno;
        const int m = allpoints[i].moveno;
        const int wn = allpoints[i].winner_moves;
        const int ln = allpoints[i].loser_moves;
        assert(wn >= ln);
        const int xc = m*N+g + 5;
        const int ylo = h+5 - yscale(wn);
        const int yhi = h+5 - yscale(ln);
        unsigned char color[3];
        game2color(g, N, color);
        printf("(%d,%d,%d,%d) => (%d,%d-%d)\n", g,m,wn,ln, xc,ylo,yhi);
        for (int j=ylo; j <= yhi; ++j) {
            assert(0 <= j && j < h);
            memcpy(im[j*w+xc], color, 3);
        }
    }
    WritePPM6(filename, im, w, h);
}

int main(int argc, char **argv)
{
    int Ngames = 0;
    pointsfile = fopen("output-points.txt", "w");
    assert(pointsfile != NULL);
    for (int i=1; i < argc; ++i) {
        FILE *in = fopen(argv[i], "r");
        if (in == NULL) {
            printf("Couldn't open file \"%s\"... skipping\n", argv[i]);
        } else {
            printf("Processing \"%s\"...\n", argv[i]);
            process(in, Ngames);
            fclose(in);
            ++Ngames;
        }
    }
    fclose(pointsfile);
    create_image("output.ppm", Ngames);
    return 0;
}


