
/* Scan a game state from standard input.
 * Find all of player 0's possible (distinct) moves for that state
 * (where two moves are "distinct" iff they result in visually
 * distinguishable game states).
 * Print the number of moves found.
 */

#include <stdio.h>
#include <vector>
#include "state.h"
#include "move.h"
#include "AllMoves.h"
#include "ApplyMove.h"
#include "AlphaBeta.h"
#include "PlanetNames.h"

typedef int Value;

struct GameStateAugmented : public GameState {
    int attacker;
    GameStateAugmented(const GameState &st, int a): GameState(st), attacker(a) { }

    static int evaluateCount, applyMoveCount, findMovesCount;

    static Value evaluate(const GameStateAugmented &sta)
    {
	++evaluateCount;
        const StarSystem *hw = sta.homeworldOf(sta.attacker);
        if (hw == NULL) return +1000;
        if (hw->ships[sta.attacker].empty()) return +1000;
        return 0;
    }
    static void applyMove(GameStateAugmented &sta, const WholeMove &move)
    {
	++applyMoveCount;
        ApplyMove::or_die(sta, sta.attacker, move);
        sta.attacker = (1 - sta.attacker);
    }
    static void findMoves(const GameStateAugmented &sta, std::vector<WholeMove> &allmoves)
    {
	++findMovesCount;
	if (sta.gameIsOver()) return;
	assert(allmoves.empty());
        findAllMoves_usualcase(sta, sta.attacker, allmoves);
        assert(!allmoves.empty());
    }
    static int findAttacker(const GameStateAugmented &sta)
    {
        return sta.attacker;
    }
};
typedef GameStateAugmented GSA;

int GSA::evaluateCount;
int GSA::applyMoveCount;
int GSA::findMovesCount;

static const char *bland_names[21] = {
    "Uuaaaa", "Uuaaab", "Uuaaac", "Uuaaad", "Uuaaae", "Uuaaaf", "Uuaaag",
    "Uuaaah", "Uuaaai", "Uuaaaj", "Uuaaak", "Uuaaal", "Uuaaam", "Uuaaao",
    "Uuaaap", "Uuaaaq", "Uuaaba", "Uuaabb", "Uuaabc", "Uuaabd", "Uuaabe"
};

int main(int argc, char **argv)
{
    if (argc != 3 || !(!strcmp(argv[1], "0") || !strcmp(argv[1], "1"))) {
        puts("./rollout <attacker> <ply>");
        puts("Given a position on standard input, find the best move for <attacker>");
        puts("from that position. Winning is better than default, which is better");
        puts("than losing; only these three conditions are distinguished.");
        exit(0);
    }
    const int attacker = atoi(argv[1]);
    const int ply = atoi(argv[2]);

    GameState st;
    st.scan(stdin);
    assignPlanetNames(st, bland_names);
    printf("----------------------\n");
    printf("%s", st.toString().c_str());
    printf("----------------------\n");
    
    GameStateAugmented sta(st, attacker);
    AlphaBeta<GameStateAugmented, WholeMove, Value> ab(GSA::evaluate,
        GSA::applyMove, GSA::findMoves, GSA::findAttacker);
    WholeMove bestmove;
    Value bestvalue;
    bool has_moves = ab.depth_first_alpha_beta(sta, ply, bestmove, bestvalue, -1000, +1000);
    assert(has_moves);
    printf("e=%d a=%d f=%d\n", GSA::evaluateCount, GSA::applyMoveCount, GSA::findMovesCount);
    printf("Best move found for player %d:\n", attacker);
    printf("(%-d) %s\n", bestvalue, bestmove.toString().c_str());
    return 0;
}

